//! A bin that calculates the mean value of all inputs.
//!
//! ```text
//!   ┌─────[mean]─────┐
//!  ⇒│input     output│⇒
//!   └────────────────┘
//! ```

use super::{
    sink_names_input, source_names_output, BinBuildEnvironment,
    BinDescription, Calculator, FetchItem, GetCalibration, Item,
    Iteration, Result, Scope, SinkBin, SinkNames, SourceBin, SourceId,
    SourceNames, SourceSinkBinDescription, WriteDotSimple, SINK_INPUT,
    SOURCE_OUTPUT,
};
use crate::{error, R64};
use indexmap::IndexSet;

static BIN_TYPE: &str = "mean";

/// A bin that calculates the mean value of all inputs.
#[derive(Debug)]
pub struct Bin {
    scope: Scope,

    source_input: Box<dyn FetchItem>,

    sum: R64,
    count: usize,

    result_output: Item,
}

impl SinkBin for Bin {}

impl SourceBin for Bin {
    fn get_source_data(&self, source: &SourceId) -> Result<Item> {
        if source.id == SOURCE_OUTPUT {
            Ok(self.result_output.clone())
        } else {
            error::MissingSourceName {
                scope: self.scope.clone(),
                name: source.id.to_string(),
                bin_type: BIN_TYPE.to_string(),
            }
            .fail()
        }
    }
}

impl Calculator for Bin {
    fn calculate(&mut self, _iteration: &Iteration) -> Result<()> {
        let input = self.source_input.fetch_item(&self.scope)?;

        if let Ok(v) = input.to_float() {
            self.sum += v;
            self.count += 1;
        };
        self.result_output = if self.count > 0 {
            Item::F64(self.sum / (self.count as f64))
        } else {
            Item::Nothing
        };
        Ok(())
    }
}

/// Description for the mean bin.
#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Description;

impl BinDescription for Description {
    type Bin = Bin;

    fn check_validity(
        &self,
        _scope: &Scope,
        _get_calibration: &mut dyn GetCalibration,
    ) -> Result<()> {
        Ok(())
    }

    fn bin_type(&self) -> &'static str {
        BIN_TYPE
    }
}

impl SinkNames for Description {
    fn sink_names(&self) -> IndexSet<String> {
        sink_names_input()
    }
}

impl SourceNames for Description {
    fn source_names(&self) -> Result<IndexSet<String>> {
        Ok(source_names_output())
    }
}

impl SourceSinkBinDescription for Description {
    fn build_bin(
        &self,
        scope: &Scope,
        env: &mut dyn BinBuildEnvironment,
    ) -> Result<Self::Bin> {
        Ok(Bin {
            scope: scope.clone(),
            source_input: env.resolve(SINK_INPUT)?,
            sum: R64::from(0f64),
            count: 0usize,
            result_output: Item::Nothing,
        })
    }
}

impl WriteDotSimple for Description {}

#[cfg(test)]
mod tests {
    use super::Description;
    use crate::bins::{directsource, verificationsink};
    use crate::Item as I;
    use crate::{run_bin, Result};
    use indexmap::indexset;

    #[test]
    fn simulate() -> Result<()> {
        use crate::Item::*;

        let input = directsource::Description {
            columns: indexset!["input".to_string()],
            rows: vec![
                vec![Nothing],
                vec![I::from(4f64)],
                vec![Nothing],
                vec![I::from(2f64)],
                vec![Nothing],
                vec![I::from(4f64)],
            ]
            .into(),
        };
        let verification = verificationsink::Description {
            columns: indexset!["output".to_string()],
            expected: vec![
                vec![Nothing],
                vec![I::from(4f64)],
                vec![I::from(4f64)],
                vec![I::from(3f64)],
                vec![I::from(3f64)],
                vec![I::from(3.33333333333333333333333f64)],
            ]
            .into(),
        };

        run_bin(&input, &Description {}, &verification)
    }
}
